import click

from .colors import CYAN
from .spell import spell_correct


class SpellcheckableAliasableGroup(click.Group):
    """
    Custom Group subclass that also has aliasing and spell correction.
    """

    def __init__(self, *args, **kwargs):
        super(SpellcheckableAliasableGroup, self).__init__(*args, **kwargs)
        self.aliases = {}

    def add_alias(self, command, alias_name):
        """
        Aliases `alias_name` to run `command`
        """
        self.aliases[alias_name] = command

    def suggest_command(self, ctx, cmd_name):
        """
        Uses edit distance to suggest a command if there's no match.
        """
        suggestion = spell_correct(cmd_name, self.list_commands(ctx) + list(self.aliases.keys()))
        if suggestion:
            ctx.fail('No such command "{cmd_name}", are you trying to run: {suggestion}?'.format(
                cmd_name=cmd_name,
                suggestion=CYAN(suggestion),
            ))

    def get_command(self, ctx, cmd_name):
        # Try to get the command directly or via an alias
        cmd = super(SpellcheckableAliasableGroup, self).get_command(ctx, cmd_name) or self.aliases.get(cmd_name)
        if cmd:
            return cmd
        # Run spellcheck
        self.suggest_command(ctx, cmd_name)



import click
import collections
import pkg_resources
import sys
import os
import traceback
import attr

from .alias_group import SpellcheckableAliasableGroup
from .colors import PURPLE, RED
from .tasks import RootTask
from ..config import Config
from ..constants import PluginHook
from ..docker.hosts import HostManager
from ..exceptions import DockerNotAvailableError
from ..containers.graph import ContainerGraph
from ..containers.profile import NullProfile, Profile
from ..utils.sorting import dependency_sort


@attr.s
class App(object):
    """
    Main app object that's passed around.

    Contains a "hooks" system, which allows plugins to register hooks (callables
    that take keyword arguments and return nothing) and other code to call them.

    Also contains a "catalog" system, which allows registration of "catalog types"
    and "catalog items", which is similar to the Python entrypoint system but tied
    to Bay plugins instead so we can have conditional loading/ordered loading.
    """
    cli = attr.ib()
    plugins = attr.ib(default=attr.Factory(dict), init=False)

    def load_config(self, config_paths):
        self.config = Config(config_paths)
        self.hosts = HostManager.from_config(self.config)
        self.containers = ContainerGraph(self.config["bay"]["home"])
        self.root_task = RootTask()

    def load_plugins(self):
        """
        Loads all plugins defined in config
        """
        self.hooks = {}
        self.catalog = {}
        # Load plugin classes based on entrypoints
        plugins = []
        for entrypoint in pkg_resources.iter_entry_points("bay.plugins"):
            try:
                plugin = entrypoint.load()
                plugins.append(plugin)
            except ImportError:
                click.echo(PURPLE("Failed to import plugin: {name}".format(name=entrypoint.name)), err=True)
                click.echo(PURPLE(traceback.format_exc()), err=True)
                sys.exit(1)
        # Build plugin provides
        provided = {}
        for plugin in plugins:
            for p in plugin.provides:
                # Make sure another plugin does not provide this
                if p in provided:
                    click.echo(PURPLE("Multiple plugins provide {}, please unload one.".format(p)))
                    sys.exit(1)
                provided[p] = plugin
        # Check plugin requires
        for plugin in plugins:
            for r in plugin.requires:
                if r not in provided:
                    click.echo(PURPLE("Plugin {} requires {}, but nothing provides it.".format(plugin, r)))
                    sys.exit(1)
        # Sort plugins by dependency order, and then alphabetically inside that
        plugins = dependency_sort(plugins, lambda x: [provided[r] for r in x.requires])
        # Load plugins
        for plugin in plugins:
            # We store plugins so you can look their instances up by class
            self.plugins[plugin] = instance = plugin(self)
            instance.load()

    def load_profiles(self):
        """
        Loads the current profile stack
        """
        self.user_profile_path = os.path.join(
            self.config["bay"]["user_profile_home"],
            self.containers.prefix,
            "user_profile.yaml"
        )

        # Load the user profile, if it exists.
        if os.path.exists(self.user_profile_path):
            self.user_profile = Profile(self.user_profile_path)
        else:
            self.user_profile = NullProfile()
        self.profiles = [self.user_profile]

        # Keep following the parent profile tree until we hit the end
        while self.profiles[-1].parent_profile:
            next_profile_path = os.path.join(
                self.config["bay"]["home"],
                "profiles",
                "{}.yaml".format(self.profiles[-1].parent_profile)
            )
            if not os.path.exists(next_profile_path):
                raise RuntimeError("Cannot load profile %s" % next_profile_path)
            self.profiles.append(Profile(
                next_profile_path,
                default_boot_compatability=True,
            ))

        # Now apply them in reverse order
        for profile in reversed(self.profiles):
            profile.apply(self.containers)

    def add_hook(self, hook_type, receiver):
        """
        Adds a plugin hook to be run later.
        """
        if hook_type not in PluginHook.valid_hooks:
            raise ValueError("Invalid hook type {}".format(hook_type))
        self.hooks.setdefault(hook_type, []).append(receiver)

    def run_hooks(self, hook_type, **kwargs):
        """
        Runs all hooks of the given type with the given keyword arguments.

        Returns True if at least one hook ran, False otherwise.
        """
        hooks = self.hooks.get(hook_type, [])
        for hook in hooks:
            hook(**kwargs)
        return bool(hooks)

    def add_catalog_type(self, name):
        """
        Adds a type of "catalog" for things to register.
        """
        if name in self.catalog:
            raise ValueError("Catalog type {} already registered".format(name))
        self.catalog[name] = collections.OrderedDict()

    def add_catalog_item(self, type_name, name, value):
        """
        Adds a catalog item by name and type
        """
        if type_name not in self.catalog:
            raise ValueError("Catalog type {} does not exist".format(type_name))
        if name in self.catalog[type_name]:
            raise ValueError("Catalog item {}/{} already registered".format(type_name, name))
        self.catalog[type_name][name] = value

    def get_catalog_items(self, type_name):
        if type_name not in self.catalog:
            raise ValueError("Catalog type {} does not exist".format(type_name))
        return self.catalog[type_name]

    def get_plugin(self, klass):
        """
        Given a plugin's class, returns the instance of it we have loaded.
        """
        return self.plugins[klass]

    def invoke(self, command_name, **kwargs):
        """
        Runs a [sub]command by name, passing context automatically.
        """
        context = click.get_current_context()
        command = cli.get_command(context, command_name)
        context.invoke(command, **kwargs)


class AppGroup(SpellcheckableAliasableGroup):
    """
    Group subclass that instantiates an App instance when called, loads
    plugins, and passes the app as the context obj.
    """

    def __init__(self, app_class, **kwargs):
        super(AppGroup, self).__init__(**kwargs)
        self.app = app_class(self)
        self.app.load_plugins()

    def invoke(self, ctx):
        ctx.obj = self.app
        return super(AppGroup, self).invoke(ctx)

    def main(self, *args, **kwargs):
        try:
            return super(AppGroup, self).main(*args, **kwargs)
        except DockerNotAvailableError as e:
            # Run the failure hooks, printing a default error if nothing is hooked in
            if not self.app.run_hooks(PluginHook.DOCKER_FAILURE):
                click.echo(RED(str(e)))
            sys.exit(1)


@click.command(cls=AppGroup, app_class=App)
@click.option('-c', '--config', multiple=True)
@click.version_option()
@click.pass_obj
def cli(app, config):
    """
    Bay, the Docker-based development environment management tool.
    """
    # Load config based on CLI parameters
    app.load_config(config)
    app.load_profiles()


# Run CLI if called directly
if __name__ == '__main__':
    cli()



class BadConfigError(Exception):
    """
    Raised when a config file is misformatted or mis-typed.
    """
    pass


class FailedCommandException(Exception):
    def __init__(self, source=None, msg=None):
        self.source = source

    def __str__(self):
        return str(self.source)

    def __repr__(self):
        return "<FailedCommandException %s>" % (self,)


class NotFoundException(FailedCommandException):
    pass


class DockerAccessException(FailedCommandException):
    """
    Raised when the docker socket is inaccessible.
    """
    def __init__(self, source, msg=''):
        super(DockerAccessException, self).__init__(source=source)
        self.msg = msg


class BuildFailureError(Exception):
    """
    Raised when a container fails to build.
    """


class DockerRuntimeError(Exception):
    """
    Raised when the Docker status cannot be correctly understood or something
    fails during the run process.
    """

    def __init__(self, message, code=None, instance=None):
        super(DockerRuntimeError, self).__init__(message)
        self.code = code
        self.instance = instance


class ImageNotFoundException(Exception):
    """
    Raised when the image requested does not exist on the docker host being
    talked to.
    """

    def __init__(self, message, image=None, image_tag=None, container=None):
        super(ImageNotFoundException, self).__init__(message)
        self.image = image
        self.image_tag = image_tag
        self.container = container


class ImagePullFailure(Exception):
    """
    Raised when the image fails to be pulled or tagged from the docker
    remote registry.
    """

    def __init__(self, message, remote_name=None, image_tag=None):
        super(ImagePullFailure, self).__init__(message)
        self.remote_name = remote_name
        self.image_tag = image_tag


class DockerInteractiveException(Exception):
    """
    Raised when a starting thread wants to run an interactive session.
    """

    def __init__(self, handler):
        self.handler = handler


class DockerNotAvailableError(Exception):
    """
    Raised when Docker is not available (the socket/machine is gone)
    """



import re
import click


WHITE = lambda s: click.style(s, fg='white')
PURPLE = lambda s: click.style(s, fg='magenta')
BLUE = lambda s: click.style(s, fg='blue')
CYAN = lambda s: click.style(s, fg='cyan')
GREEN = lambda s: click.style(s, fg='green')
YELLOW = lambda s: click.style(s, fg='yellow')
RED = lambda s: click.style(s, fg='red')
BOLD = lambda s: click.style(s, bold=True)
FADED = lambda s: click.style(s, dim=True)

REPLACE_LINE = lambda: click.echo('\033[F\r\033[K', nl=False)
ERASE_LINE = lambda: click.echo('\r\033[K', nl=False)


def yesno(value):
    """
    Converts and colors True/False values into "Yes/No"
    """
    if value:
        return GREEN("Yes")
    else:
        return RED("No")


def remove_ansi(line):
    """
    Removes ANSI control characters from a string
    """
    ansi_escape = re.compile(r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]')
    return ansi_escape.sub('', line)



import os
import warnings
import yaml

import attr

from ..exceptions import BadConfigError


@attr.s
class Profile:
    """
    Represents a profile - a way of running containers based on a base graph.

    A profile comes from a single config file, and then applies changes from that
    to a ContainerGraph. Multiple profiles might be used one after the other.
    """
    file_path = attr.ib(default=None)
    load_immediately = attr.ib(default=True)
    default_boot_compatability = attr.ib(default=False)
    parent_profile = attr.ib(default=None, init=False)
    description = attr.ib(default=None, init=False)
    version = attr.ib(default=None, init=False)
    containers = attr.ib(default=attr.Factory(dict), init=False)

    def __attrs_post_init__(self):
        if self.load_immediately:
            self.load()
        self.name = os.path.basename(self.file_path).split(".")[0]

    def load(self):
        """
        Loads the profile data from a YAML file
        """
        # Read in file
        with open(self.file_path, "r") as fh:
            data = yaml.safe_load(fh.read())
        if data is None:
            return

        # Parse container details
        try:
            self.parent_profile = data.get("inherits", data.get("name"))
        except AttributeError:
            self.parent_profile = None  # The parent profile is a null.

        self.description = data.get("description")
        self.version = data.get("min-version")

        for name, details in data.get('containers', {}).items():
            if details is None:
                details = {}
            self.containers[name] = {
                "links": details.get("links") or {},
                "devmodes": set(details.get("devmodes") or []),
                "ports": details.get("ports") or {},
                "environment": details.get("environment") or {},
                "ephemeral": details.get("ephemeral") or False,
            }
            # Make sure links has dicts for the right things
            self.containers[name]["links"].setdefault("optional", [])
            self.containers[name]["links"].setdefault("required", [])
            # Merge legacy settings into those
            # TODO: Remove old profile links format
            self.containers[name]["links"]["optional"].extend(details.get("ignore_links") or [])
            self.containers[name]["links"]["required"].extend(details.get("extra_links") or [])

    def dump(self):
        data = {
            "name": self.parent_profile,
        }
        containers = {}

        for container_name, container_data in self.containers.items():
            if container_data and not container_data.get('ephemeral'):
                container_details_to_write = {}
                for k, v in container_data.items():
                    # Only write out links if populated
                    if k == "links":
                        if v["optional"]:
                            container_details_to_write.setdefault("links", {})
                            container_details_to_write["links"]["optional"] = sorted(v["optional"])
                        if v["required"]:
                            container_details_to_write.setdefault("links", {})
                            container_details_to_write["links"]["required"] = sorted(v["required"])
                    # Serialize sets as sorted lists
                    elif v:
                        if isinstance(v, set):
                            container_details_to_write[k] = sorted(v)
                        else:
                            container_details_to_write[k] = v

                if container_details_to_write:
                    containers[container_name] = container_details_to_write

        if containers:
            data["containers"] = containers

        if self.version:
            data['min-version'] = self.version

        return data

    def apply(self, graph):
        """
        Applies the profile to the given graph
        """
        self.graph = graph
        for name, details in self.containers.items():
            try:
                container = self.graph[name]
            except KeyError:
                warnings.warn("Cannot apply profile for nonexistent container {}".format(name))
                continue
            # Apply container links
            if "links" in details:
                if details["links"]["required"] or details["links"]["optional"]:
                    self.graph.set_dependencies(
                        container,
                        [self.graph[link]
                        for link in self.calculate_links(container)],
                    )
            # Set flag saying it's specified in a profile (for bay build
            # profile) - not set for the user profile for now
            # TODO: remove user profile restriction with default boot compat stuff
            if self.default_boot_compatability:
                self.graph.set_option(container, "in_profile", True)
            # Set default boot mode
            if details.get('default_boot'):
                self.graph.set_option(container, "default_boot", True)
            # TODO: Remove this temporary fix that allows parent profiles
            # default boot based on just having the container in the profile
            # (provided it is not a foreground container)
            if not container.foreground and self.default_boot_compatability:
                self.graph.set_option(container, "default_boot", True)
            # Set devmodes
            self.graph.set_option(container, "devmodes", details["devmodes"])
            # Set ports to apply
            if "ports" in details:
                for a, b in details["ports"].items():
                    try:
                        container.ports[int(a)] = int(b)
                    except TypeError:
                        raise BadConfigError('Profile contains invalid ports for {}: {}'.format(a, b))
            # Apply any image tag override
            if "image_tag" in details:
                container.image_tag = details['image_tag']
            # Store environment variables
            for key, value in details.get('environment', {}).items():
                container.environment[key] = value

    def calculate_links(self, container):
        """
        Works out what links the container should have
        """
        # Check that they are all valid links
        optional_links = self.containers[container.name]['links']['optional']
        required_links = self.containers[container.name]['links']['required']
        for link_name in optional_links:
            if link_name not in container.links:
                raise BadConfigError(
                    "Profile contains invalid optional link for {}: {}".format(container.name, link_name)
                )
        for link_name in required_links:
            if link_name not in container.links:
                raise BadConfigError(
                    "Profile contains invalid required link for {}: {}".format(container.name, link_name)
                )
        # Work out desired final set of links
        current_dependencies = [c.name for c in self.graph.dependencies(container)]
        return {
            link_name
            for link_name, link_options in container.links.items()
            if (
                (link_name in current_dependencies and link_name not in optional_links) or
                (link_name in required_links)
            )
        }

    def save(self):
        """
        Saves the user profile things to disc after loading.

        Persists the profile to disk as YAML
        """
        # Set user profile to ~/.bay/eventbrite/user_profile.yaml
        try:
            os.makedirs(os.path.dirname(self.file_path))
        except OSError:
            pass
        with open(self.file_path, "w") as fh:
            yaml.safe_dump(self.dump(), fh, default_flow_style=False, indent=4)


@attr.s
class NullProfile(Profile):
    file_path = attr.ib(default=None, init=False)

    def load(self):
        """
        Loads the empty profile.
        """
        self.containers = {}

    def save(self):
        """
        Raises an error, you can't save a NullProfile.
        """
        raise BadConfigError("You can't save a NullProfile, please load a profile using `bay profile <profile_name>`")

    def calculate_links(self, container):
        """
        Returns None, a NullProfile has no links.
        """

    def apply(self, graph):
        """
        Returns None, you can't apply a NullProfile to a container-graph.
        """



import os
import yaml

import attr

from ..exceptions import BadConfigError
from .container import Container


@attr.s
class ContainerGraph:
    """
    Represents the graph of all available containers.

    Containers are loaded from a source directory. The graph has containers as
    nodes and (runtime) dependencies as edges. It also stores a set of "options"
    for each container, which contain information like which devmodes to apply/
    containers to start by default.
    """
    path = attr.ib(convert=os.path.abspath)
    containers = attr.ib(default=attr.Factory(dict), init=False, repr=False)
    _dependencies = attr.ib(default=attr.Factory(dict), repr=False)
    _build_dependencies = attr.ib(default=attr.Factory(dict), init=False, repr=False)
    _options = attr.ib(default=attr.Factory(dict), init=False, repr=False)
    config_path = attr.ib(init=False)

    def __attrs_post_init__(self):
        """
        Loads containers from the filesystem and sets up the graph.
        """
        self.load_config()
        self.load_containers()

    def load_config(self):
        """
        Loads top-level configuration
        """
        self.prefix = None
        self.registry = None
        # Work out the path to the configuration file
        self.config_path = os.path.join(self.path, "bay.yaml")
        if not os.path.isfile(self.config_path):
            self.config_path = os.path.join(self.path, "tug.yaml")
            if not os.path.isfile(self.config_path):
                raise BadConfigError("Cannot find bay.yaml in top level of container library at %s" % self.path)
        # Load the configuration file
        with open(self.config_path, "r") as fh:
            config_data = yaml.safe_load(fh.read())
        if not isinstance(config_data, dict):
            raise BadConfigError("%s is badly formatted (not a dict)" % self.config_path)
        for key, value in config_data.items():
            if key == "prefix":
                self.prefix = value
            elif key == "registry":
                self.registry = value
            else:
                raise BadConfigError("Unknown key in %s: %s" % (self.config_path, key))
        if self.prefix is None:
            raise BadConfigError("No prefix set in top-level tug.yaml {}".format(self.config_path))

    def load_containers(self):
        """
        Loads containers from their directories.
        """
        # Scan through directories and load ones that look right
        containers = []
        for name in os.listdir(self.path):
            container_path = os.path.join(self.path, name)
            if os.path.isdir(container_path) and os.path.isfile(os.path.join(container_path, "Dockerfile")):
                containers.extend(Container.from_directory(self, container_path))
        self.add_containers(containers)

    def add_containers(self, containers):
        """
        Adds container nodes to the graph and links them up based on their `links`
        """
        # Add to main set
        for container in containers:
            self.containers[container.name] = container
        # Link by dependency
        for container in containers:
            # Runtime dependencies
            self.set_dependencies(
                container,
                [
                    self.containers[link_name]
                    for link_name, link_details
                    in container.links.items()
                    if link_details.get("required")
                ],
            )
            # Build dependencies
            if container.build_parent_in_prefix:
                self.add_build_dependency(container, self.containers[container.build_parent.split("/")[1]])

    def set_dependencies(self, depender, providers):
        """
        Adds runtime dependency edges to the graph where `depender` depends on each of `providers`
        """
        self._dependencies[depender] = set()
        for provider in providers:
            if provider not in self.containers.values() or depender not in self.containers.values():
                raise ValueError("Cannot link between containers %s and %s - one or both not in graph." % (
                    provider,
                    depender,
                ))
            self._dependencies[depender].add(provider)

    def set_option(self, container, option, value):
        """
        Sets the option named `option` on the container to the given value.
        Also does some basic checks of what valid options are.
        """
        if option in ("default_boot", "in_profile"):
            value = bool(value)
        elif option == "devmodes":
            if not isinstance(value, set):
                raise BadConfigError("Devmodes option must be a set")
        else:
            raise BadConfigError("Unknown option %s being set" % option)
        self._options.setdefault(container, {})[option] = value

    def add_build_dependency(self, depender, provider):
        """
        Adds a build dependency edge to the graph where `depender` depends on `provider`
        """
        if provider not in self.containers.values() or depender not in self.containers.values():
            raise ValueError("Cannot build-link between containers %s and %s - one or both not in graph." % (
                provider,
                depender,
            ))
        self._build_dependencies[depender] = provider

    def dependencies(self, container):
        """
        Returns a set of the dependencies of the container
        """
        return self._dependencies.get(container, set())

    def dependents(self, container):
        """
        Returns the containers that depend on the named container
        """
        result = set()
        for candidate, dependencies in self._dependencies.items():
            if container in dependencies:
                result.add(candidate)
        return result

    def build_ancestry(self, container):
        """
        Returns the ancestors of the container based on build dependencies, in order from furthest ancestor
        to immediate parent.
        """
        ancestry = []
        while container is not None:
            ancestry.insert(0, container)
            container = self._build_dependencies.get(container, None)
        return ancestry[:-1]

    def build_parent(self, container):
        """
        Returns the immediate parent of a container, per its build dependencies.
        """
        return self._build_dependencies.get(container, None)

    def options(self, container):
        """
        Returns runtime options for the container
        """
        return self._options.get(container, {})

    def __getitem__(self, key):
        return self.containers[key]

    def __iter__(self):
        return iter(self.containers.values())



import attr
import docker
import os
import urllib.parse

from ..exceptions import DockerNotAvailableError
from ..utils.functional import cached_property, thread_cached_property
from .images import ImageRepository


@attr.s
class HostManager(object):
    """
    Contains all known hosts.
    """
    _hosts = attr.ib(default=attr.Factory(list))
    hosts = attr.ib(default=attr.Factory(dict), init=False)

    def __attrs_post_init__(self):
        for host in self._hosts:
            self.add_host(host)

    @classmethod
    def from_config(cls, config):
        return cls([
            Host.from_env(),
        ])

    def add_host(self, host):
        if host.alias in self.hosts:
            raise ValueError("Host alias %s is already assigned" % host.alias)
        self.hosts[host.alias] = host

    def __iter__(self):
        return iter(self.hosts.values())

    def __getitem__(self, key):
        return self.hosts[key]


@attr.s
class Host(object):
    """
    A Docker-running host.
    """
    alias = attr.ib()
    url = attr.ib()
    tls_ca = attr.ib()
    tls_cert = attr.ib()
    tls_key = attr.ib()
    url_scheme = attr.ib(init=False)
    url_location = attr.ib(init=False)

    def __attrs_post_init__(self):
        # Parse URL into components
        parse_result = urllib.parse.urlparse(self.url)
        self.url_scheme = parse_result.scheme
        self.url_location = parse_result.netloc
        if self.url_scheme not in ["unix", "tcp"]:
            raise ValueError("Unknown scheme in Docker URL %s" % self.url)

    @classmethod
    def from_env(cls, alias="default"):
        """
        Makes a host from Docker environment variables.
        """
        tls_ca = tls_cert = tls_key = None
        if "DOCKER_CERT_PATH" in os.environ:
            tls_ca = os.path.join(os.environ['DOCKER_CERT_PATH'], "ca.pem")
            tls_cert = os.path.join(os.environ['DOCKER_CERT_PATH'], "cert.pem")
            tls_key = os.path.join(os.environ['DOCKER_CERT_PATH'], "key.pem")
        return cls(
            alias=alias,
            url=os.environ.get("DOCKER_HOST", "unix:///var/run/docker.sock"),
            tls_ca=tls_ca,
            tls_cert=tls_cert,
            tls_key=tls_key,
        )

    @cached_property
    def publicly_visible(self):
        """
        Says if the server can be seen by other servers or not.
        """
        return (
            self.url_scheme == "tcp" and
            # TODO: Resolve hostnames first?
            self.url_location.split(".")[0] not in ["10", "192", "127"]
        )

    @cached_property
    def external_host_address(self):
        """
        Returns the address of the host as seen from an external networks and/or
        a host computer where bay is running against a VM. This is the address which
        exposed ports would apparently appear on.
        """
        if self.url_scheme == "unix":
            return "127.0.0.1"
        elif self.url_scheme == "tcp":
            return self.url_location.split(":", 1)[0]

    @cached_property
    def allow_ssh_agent(self):
        """
        Says if the server is non-shared and thus safe to run ssh-agent on
        (ssh-agent contains the SSH keys of the person who started it in
        plaintext, so it must not be used on a public server)
        """
        return not self.publicly_visible

    @thread_cached_property
    def client(self):
        """
        Returns a Docker client for the URL
        """
        # TLS setup
        tls = None
        tls_client = None
        if self.tls_cert and self.tls_key:
            tls_client = (self.tls_cert, self.tls_key)
        if tls_client or self.tls_ca:
            tls = docker.tls.TLSConfig(
                ca_cert=self.tls_ca,
                client_cert=tls_client,
                verify=True,
            )
        # Make client
        try:
            return docker.Client(
                base_url=self.url,
                version="auto",
                timeout=10,
                tls=tls,
            )
        except docker.errors.DockerException:
            raise DockerNotAvailableError("The docker host at {} is not available".format(self.url))

    @thread_cached_property
    def images(self):
        """
        Returns an image repository for the host
        """
        return ImageRepository(self)

    def container_exists(self, name):
        """
        Shortcut to see if a container exists with the given runtime name
        """
        try:
            self.client.inspect_container(name)
            return True
        except docker.errors.APIError:
            return False

    def container_running(self, name, ignore_exists=False):
        """
        Says if the named container is running or not. Errors if you provide
        a container that does not exist.
        """
        if ignore_exists and not self.container_exists(name):
            return False
        data = self.client.inspect_container(name)
        return data['State']['Running']

    @cached_property
    def build_host_ip(self):
        """
        Returns the internal IP of the host as seen from the containers during
        build, which is the gateway of the default bridge network.
        """
        return self.client.inspect_network("bridge")['IPAM']['Config'][0]['Gateway']



import contextlib
import shutil
import threading
import time

from .colors import CYAN, GREEN, RED, YELLOW
from ..utils.threading import ExceptionalThread


UP_ONE = "\033[A\033[1000D"
CLEAR_LINE = "\033[2K"

console_lock = threading.Lock()


class Task:
    """
    Something that can be started (by being created), have progress reported, and then finished.
    It can also have a number of sub-tasks, and arbitary lines of extra information
    that should be shown to the user alongside progress messages or a progress bar.
    """

    INDENT_AMOUNT = 2

    FLAVOR_NEUTRAL = "neutral"
    FLAVOR_GOOD = "good"
    FLAVOR_BAD = "bad"
    FLAVOR_WARNING = "warning"

    def __init__(self, name, parent=None, hide_if_empty=False, collapse_if_finished=False, progress_formatter=None):
        self.name = name
        # If this task only displays if it has children
        self.hide_if_empty = hide_if_empty
        # If this task collapses to just the first line if it's finished
        self.collapse_if_finished = collapse_if_finished
        # Any parent tasks to trigger updates in
        self.parent = parent
        with console_lock:
            if self.parent is not None:
                self.parent.subtasks.append(self)
        # Sub tasks to show under this one
        self.subtasks = []
        # The current status message
        self.status = None
        # The current progress from 0 - 1
        self.progress = None
        # The way to format the progress numbers
        self.progress_formatter = progress_formatter or str
        # The current status flavor (turns into a colour)
        self.status_flavor = self.FLAVOR_NEUTRAL
        # Extra lines of information to show underneath the task
        self.extra_info = []
        # If the task is complete
        self.finished = False
        # Number of lines we had previously cleared
        self.cleared_lines = 0
        # If the output is currently "paused" for other things to write to the console
        self.output_paused = False
        # Run update
        self.update()

    def update(self, status=None, status_flavor=None, progress=None, force=False):
        """
        Update either the status message, the progress bar, or both.
        If this is the topmost task, this will trigger a reprint on the console.
        """
        if self.finished and not force:
            raise ValueError("You cannot update() a finished task!")
        with console_lock:
            if status is not None:
                self.status = status
            if progress is not None:
                if len(progress) != 2:
                    raise ValueError("Progress must be a 2-tuple of (count, total)")
                self.progress = progress
            if status_flavor is not None:
                self.status_flavor = status_flavor
        # Look for a parent to potentially trigger update on, or print ourselves
        # if there isn't one
        if self.parent is not None:
            self.parent.update()
        else:
            self.clear_and_output()

    def add_extra_info(self, message):
        """
        Adds a line of extra info and triggers updates
        """
        with console_lock:
            self.extra_info.append(message)
        if self.parent is not None:
            self.parent.update()

    def set_extra_info(self, messages):
        """
        Sets all extra info and triggers updates
        """
        with console_lock:
            self.extra_info = messages
        if self.parent is not None:
            self.parent.update()

    def finish(self, **kwargs):
        """
        Marks the task as finished, meaning it can no longer be mutated.
        Used to optimise terminal output only.
        """
        self.finished = True
        self.update(force=True, **kwargs)

    def wrapped_extra_info(self, text_width):
        """
        Returns extra_info wrapped to fit the terminal width.
        """
        actual_output = []
        for line in self.extra_info:
            line = line.strip()
            while line:
                actual_output.append(line[:text_width])
                line = line[text_width:]
        return actual_output

    def make_progress_bar(self, count, total, width=30):
        """
        Helper for making progress bar text.
        """
        progress = min(max(count / total, 0), 1)
        bar_width = width - 2
        bar_size = int(bar_width * progress)
        return "[{}{}] {}/{}".format(
            "=" * bar_size,
            " " * (bar_width - bar_size),
            self.progress_formatter(count),
            self.progress_formatter(total),
        )

    def output(self, terminal_width, indent=0):
        """
        Returns the lines to output for this task to the screen (as a generator)
        """
        if self.hide_if_empty and not self.subtasks:
            return
        # Work out progress text
        progress_string = ""
        if self.progress is not None:
            progress_string = self.make_progress_bar(*self.progress) + " "
        # Work out status text
        status_string = self.status or ""
        if self.status_flavor == self.FLAVOR_BAD:
            status_string = RED(status_string)
        elif self.status_flavor == self.FLAVOR_GOOD:
            status_string = GREEN(status_string)
        elif self.status_flavor == self.FLAVOR_WARNING:
            status_string = YELLOW(status_string)
        # Print out our line
        indent_string = " " * (self.INDENT_AMOUNT * indent)
        yield "{}{}: {}{}".format(
            indent_string,
            CYAN(self.name),
            progress_string,
            status_string,
        )
        if not (self.finished and self.collapse_if_finished):
            # Print out extra info
            indent_string = (indent + 1) * (" " * self.INDENT_AMOUNT)
            for info in self.wrapped_extra_info(terminal_width - len(indent_string)):
                yield indent_string + info[:terminal_width - len(indent_string)].replace("\n", "")
            # Print out subtasks
            for subtask in self.subtasks:
                yield from subtask.output(terminal_width, indent=indent + 1)

    def clear_and_output(self):
        """
        Clears the terminal up to the right line then outputs the information
        of the task.
        """
        # See if output is paused
        if self.output_paused:
            return
        # OK, print
        with console_lock:
            # Get terminal width
            terminal_width = shutil.get_terminal_size((80, 20)).columns
            # Get the output we need to print
            output = list(self.output(terminal_width))
            # Scroll the terminal down/up enough for any new lines
            needed_lines = len(output)
            new_lines = needed_lines - self.cleared_lines
            if new_lines > 0:
                print("\n" * new_lines, flush=True, end="")
            elif new_lines < 0:
                print(
                    (UP_ONE + CLEAR_LINE) * abs(new_lines),
                    flush=True,
                    end="",
                )
            self.cleared_lines = needed_lines
            # Move cursor to top of cleared section
            print(
                (UP_ONE + CLEAR_LINE) * needed_lines,
                flush=True,
                end="",
            )
            for line in output:
                print(line)

    def _pause_output(self, pause=True):
        """
        Allows the output to be paused and unpaused by finding the parent and doing it there.
        """
        if self.parent is None:
            self.output_paused = pause
            if not pause:
                # Make the output rewrite from where it is
                self.cleared_lines = 0
                self.clear_and_output()
        else:
            self.parent._pause_output(pause)

    @contextlib.contextmanager
    def paused_output(self):
        """
        Context manager that pauses printing of output until it's exited.
        """
        self._pause_output(True)
        yield
        self._pause_output(False)

    @contextlib.contextmanager
    def rate_limit(self, interval=0.1):
        """
        Context manager that rate-limits updates on tasks
        """
        buffered_changes = {"running": True}

        # Thread loop that flushes every interval
        def flusher():
            while buffered_changes['running']:
                # Do any extra_info calls
                if "set_extra_info" in buffered_changes:
                    self.set_extra_info(buffered_changes['set_extra_info'])
                    del buffered_changes['set_extra_info']
                # Do any update calls
                if "update" in buffered_changes:
                    self.update(**buffered_changes['update'])
                    del buffered_changes['update']
                # Sleep
                time.sleep(interval)

        # Fake task object to provide out
        class BufferedTask(object):

            def set_extra_info(self, extra_info):
                self.buffered_changes['set_extra_info'] = extra_info

            def update(self, **kwargs):
                self.buffered_changes['update'] = kwargs

        # Start thread that flushes every interval
        flush_thread = ExceptionalThread(target=flusher, daemon=True)
        flush_thread.start()

        # Run inner code
        yield BufferedTask()

        # Do one more flush and exit
        buffered_changes['running'] = False
        flush_thread.join()


class RootTask(Task):
    """
    Special task subclass that represents the "root" task, the instance that
    has no output of its own but encapsulates all other tasks in the app in order.
    """

    def __init__(self):
        super(RootTask, self).__init__("__root__")

    def output(self, terminal_width):
        for subtask in self.subtasks:
            yield from subtask.output(terminal_width, indent=0)



import attr


class PluginMetaclass(type):
    """
    Custom plugin metaclass to allow comparison for sorting by memory ID
    """

    @property
    def _orderable_id(self):
        """
        Property that is unique but orderable enough that plugins load
        in the same order each call unless they're named exactly the same.
        """
        return self.__name__ + str(id(self))

    def __gt__(self, other):
        return self._orderable_id > other._orderable_id

    def __lt__(self, other):
        return self._orderable_id < other._orderable_id

    def __eq__(self, other):
        return self._orderable_id == other._orderable_id

    def __ne__(self, other):
        return self._orderable_id != other._orderable_id

    def __hash__(self):
        return hash(self._orderable_id)


@attr.s
class BasePlugin(object, metaclass=PluginMetaclass):
    """
    Base plugin template.
    """
    app = attr.ib()

    # Simple plugin dependency checking - any strings in requires must be
    # in exactly one other loaded plugin in provides.
    provides = []
    requires = []

    def load(self):
        pass

    def add_command(self, func):
        self.app.cli.add_command(func)

    def add_alias(self, command_name, alias):
        self.app.cli.add_alias(command_name, alias)

    def add_hook(self, hook_type, func):
        self.app.add_hook(hook_type, func)

    def add_catalog_type(self, name):
        self.app.add_catalog_type(name)

    def add_catalog_item(self, type_name, name, value):
        self.app.add_catalog_item(type_name, name, value)



import attr
import click

from .colors import CYAN


@attr.s
class Table(object):
    """
    Table pretty-printer.

    Initialise with a list of ("COLUMNNAME", width_int). One space is added
    between columns.
    """
    columns = attr.ib()
    format_string = attr.ib(init=False)

    def __attrs_post_init__(self):
        self.format_string = " ".join(
            "%%-%is" % width
            for name, width in self.columns
        )

    def print_header(self):
        click.echo(CYAN(self.format_string % tuple(name for name, width in self.columns)))

    def print_row(self, data):
        click.echo(self.format_string % tuple(data))

    def print_all(self, datas):
        """
        Prints a table with header and all data at once
        """
        self.print_header()
        for data in datas:
            self.print_data(data)



class PluginHook:
    PRE_BUILD = "pre-build"
    POST_BUILD = "post-build"
    PRE_START = "pre-start"
    POST_START = "post-start"
    DOCKER_FAILURE = "docker-fail"

    valid_hooks = frozenset([
        PRE_BUILD,
        POST_BUILD,
        PRE_START,
        POST_START,
        DOCKER_FAILURE,
    ])



import attr
from docker.errors import NotFound

from .base import BasePlugin
from ..cli.tasks import Task
from ..constants import PluginHook
from ..docker.build import Builder


@attr.s
class BuildVolumesPlugin(BasePlugin):
    """
    Plugin for showing information about containers
    """

    requires = ["build"]

    def load(self):
        self.add_hook(PluginHook.PRE_START, self.pre_start)
        self.add_hook(PluginHook.POST_BUILD, self.post_build)

    def pre_start(self, host, instance, task):
        """
        Safety net to stop you booting volume-providing containers normally,
        and to catch and build volume containers if they're needed
        """
        # Safety net
        if instance.container.extra_data.get("provides-volume", None):
            raise ValueError("You cannot run a volume-providing container {}".format(instance.container.name))
        # If the container has named volumes, see if they're provided by anything else
        # and if so, if they're built.
        # First, collect what volumes are provided by what containers
        providers = {}
        for container in self.app.containers:
            provides_volume = container.extra_data.get("provides-volume", None)
            if provides_volume:
                providers[provides_volume] = container
        # Now see if any of the volumes we're trying to add need it
        for _, name in instance.container.named_volumes.items():
            if name in providers:
                # Alright, this is one that could be provided. Does it already exist?
                try:
                    host.client.inspect_volume(name)
                except NotFound:
                    # Aha! Build it!
                    Builder(
                        host,
                        providers[name],
                        self.app,
                        parent_task=task,
                        logfile_name=self.app.config.get_path(
                            'bay',
                            'build_log_path',
                            self.app,
                        ),
                        verbose=True,
                    ).build()

    def post_build(self, host, container, task):
        """
        Intercepts builds of volume-providing containers and unpacks them.
        """
        provides_volume = container.extra_data.get("provides-volume", None)
        if provides_volume:
            volume_task = Task("Extracting into volume {}".format(provides_volume), parent=task)
            # Configure the container
            volume_mountpoints = ["/volume/"]
            volume_binds = {provides_volume: {"bind": "/volume/", "mode": "rw"}}
            container_pointer = host.client.create_container(
                container.image_name,
                detach=False,
                volumes=volume_mountpoints,
                host_config=host.client.create_host_config(
                    binds=volume_binds,
                ),
            )
            # Start it in the foreground so we wait till it exits (detach=False above)
            volume_task.update(status="Extracting")
            host.client.start(container_pointer)
            host.client.wait(container_pointer['Id'])
            host.client.remove_container(container_pointer['Id'])
            volume_task.update(status="Done", status_flavor=Task.FLAVOR_GOOD)



import attr
import http.client
import socket
import time
from docker.errors import NotFound

from .base import BasePlugin
from ..cli.tasks import Task
from ..constants import PluginHook
from ..exceptions import DockerRuntimeError


class WaitsPlugin(BasePlugin):
    """
    Contains the basic, standard waits. Waits' .check is called repeatedly and should return True if the condition is
    met or False if it is not.
    """

    provides = ["waits"]

    def load(self):
        self.add_hook(PluginHook.POST_START, self.post_start)
        self.add_catalog_type("wait")
        self.add_catalog_item("wait", "http", HttpWait)
        self.add_catalog_item("wait", "https", HttpsWait)
        self.add_catalog_item("wait", "tcp", TcpWait)
        self.add_catalog_item("wait", "time", TimeWait)
        self.add_catalog_item("wait", "file", FileWait)

    def post_start(self, host, instance, task):
        # Loop through all waits and build instances
        wait_instances = []
        for wait in instance.container.waits:
            # Look up wait in app
            try:
                wait_class = self.app.get_catalog_items("wait")[wait["type"]]
            except KeyError:
                raise DockerRuntimeError(
                    "Unknown wait type {} for {}".format(wait["type"], instance.container.name)
                )
            # Initialise it and attach a task
            params = wait.get("params", {})
            params["instance"] = instance
            params["host"] = host
            wait_instance = wait_class(**params)
            wait_instance.task = Task("Waiting for {}".format(wait_instance.description()), parent=task)
            wait_instance.task.update(status="Waiting")
            wait_instances.append(wait_instance)

        # Check on them all until they finish
        while wait_instances:
            # See if the container actually died
            if not host.container_running(instance.name):
                task.update(status="Dead", status_flavor=Task.FLAVOR_BAD)
                raise DockerRuntimeError(
                    "Container {} died while waiting for boot completion".format(instance.container.name)
                )
            # Check the waits
            for wait_instance in list(wait_instances):
                if wait_instance.ready():
                    wait_instance.task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)
                    wait_instances.remove(wait_instance)
            time.sleep(1)


@attr.s
class TcpWait:
    """
    Checks that a TCP port is open
    """

    instance = attr.ib()
    host = attr.ib()
    port = attr.ib(default=80)
    timeout = attr.ib(default=1)

    def ready(self):
        try:
            conn_kwargs = {}
            if self.timeout:
                conn_kwargs['timeout'] = self.timeout
            if self.port not in self.instance.port_mapping:
                raise DockerRuntimeError("Trying to wait on non-exposed port {}".format(self.port))
            conn = socket.create_connection(self.target(), **conn_kwargs)
            conn.close()
            return True
        except socket.error:
            return False

    def target(self):
        """
        Returns (host, port) target information.
        """
        if self.port not in self.instance.port_mapping:
            raise DockerRuntimeError("Trying to wait on non-exposed port {}".format(self.port))
        return (self.host.external_host_address, self.instance.port_mapping[self.port])

    def description(self):
        return "TCP on port {}".format(self.port)


@attr.s
class HttpWait(TcpWait):
    """
    Checks that a HTTP endpoint exists and returns a good value.
    """

    path = attr.ib(default="/")
    method = attr.ib(default="GET")
    headers = attr.ib(default=attr.Factory(dict))
    expected_codes = attr.ib(default=attr.Factory(lambda: range(200, 400)))

    connection_class = http.client.HTTPConnection

    def ready(self):
        addr, port = self.target()
        conn = self.connection_class(addr, port, timeout=self.timeout)
        # Run wait
        try:
            conn.request(self.method, self.path, headers=self.headers)
            response = conn.getresponse()
            if response.status in self.expected_codes:
                return True
        except:
            return False
        finally:
            conn.close()

    def description(self):
        return "HTTP on port {}".format(self.port)


class HttpsWait(HttpWait):
    """
    HTTPS variant of the HTTP wait
    """
    connection_class = http.client.HTTPSConnection

    def description(self):
        return "HTTPS on port {}".format(self.port)


@attr.s
class TimeWait:
    """
    Waits a number of seconds
    """

    # Everything needs instance and host, alas
    instance = attr.ib()
    host = attr.ib()
    seconds = attr.ib()

    def __attrs_post_init__(self):
        self.wait_until = time.time() + int(self.seconds)

    def ready(self):
        return time.time() >= self.wait_until

    def description(self):
        return "{} seconds".format(self.seconds)


@attr.s
class FileWait:
    """
    Waits until a named file appears in the container
    """

    instance = attr.ib()
    host = attr.ib()
    path = attr.ib()
    waiting_name = attr.ib(default=None)

    def ready(self):
        # Get the file from the host
        try:
            self.host.client.get_archive(self.instance.name, self.path)
        except NotFound:
            return False
        else:
            return True

    def description(self):
        return self.waiting_name or "file {}".format(self.path)



import re
import os
import yaml

import attr

from ..exceptions import BadConfigError


@attr.s
class Container:
    """
    Represents a single container type that's available to run (not a running container -
    we call those Instances).

    All containers are backed by a local disk directory containing their information,
    even if the actual running server is remote.
    """
    parent_pattern = re.compile(r'^FROM\s+([\S/]+)', re.IGNORECASE)
    git_volume_pattern = re.compile(r'^\{git@github.com:eventbrite/([\w\s\-]+).git\}(.*)$')

    graph = attr.ib(repr=False, hash=False, cmp=False)
    path = attr.ib(repr=False, hash=True, cmp=True)
    suffix = attr.ib(repr=False, hash=False, cmp=False)
    dockerfile_name = attr.ib(repr=False, hash=False, cmp=False)
    name = attr.ib(init=False, repr=True, hash=True, cmp=True)

    def __attrs_post_init__(self):
        self.load()

    @classmethod
    def from_directory(cls, graph, path):
        """
        Creates a set of one or more Container objects from a source directory.
        The "versions" key in the bay.yaml file can mean there are multiple variants
        of the container, and we treat each separately (though they share a build directory and bay.yaml settings)
        """
        versions = {None: "Dockerfile"}
        # Read config and get out any non-default versions
        config_path = os.path.join(path, "bay.yaml")
        if not os.path.isfile(config_path):
            config_path = os.path.join(path, "tug.yaml")
        if os.path.isfile(config_path):
            # Read out config, making sure a empty file (None) appears as empty dict
            with open(config_path, "r") as fh:
                config_data = yaml.safe_load(fh.read()) or {}
                # Merges extra versions in config file into versions dict
                versions.update({
                    str(suffix): dockerfile_name
                    for suffix, dockerfile_name in config_data.get("versions", {}).items()
                })
        # For each version, make a Container class for it, and return the list of them
        return [
            cls(graph, path, suffix, dockerfile_name)
            for suffix, dockerfile_name in versions.items()
        ]

    def load(self):
        """
        Loads information from the container's files.
        """
        # Work out paths to key files, make sure they exist
        self.dockerfile_path = os.path.join(self.path, self.dockerfile_name)
        self.config_path = os.path.join(self.path, "bay.yaml")
        if not os.path.isfile(self.config_path):
            self.config_path = os.path.join(self.path, "tug.yaml")
        if not os.path.isfile(self.dockerfile_path):
            raise BadConfigError("Cannot find Dockerfile for container %s" % self.path)
        # Calculate name from path component
        if self.suffix is None:
            self.name = os.path.basename(self.path)
        else:
            self.name = os.path.basename(self.path) + "-" + self.suffix
        self.image_name = '{prefix}/{name}'.format(
            prefix=self.graph.prefix,
            name=self.name,
        )
        # Load parent image and possible build args from Dockerfile
        self.possible_buildargs = set()
        self.build_parent = None
        with open(self.dockerfile_path, "r") as fh:
            for line in fh:
                parent_match = self.parent_pattern.match(line)
                if parent_match:
                    self.build_parent = parent_match.group(1)
                    # Make sure any ":" in the parent is changed to a "-"
                    # TODO: Add warning here once we've converted enough of the dockerfiles
                    self.build_parent = self.build_parent.replace(":", "-")
                elif line.lower().startswith("arg "):
                    self.possible_buildargs.add(line.split()[1])
        if self.build_parent is None:
            raise BadConfigError("Container {} has no valid FROM line".format(self.path))
        self.build_parent_in_prefix = self.build_parent.startswith(self.graph.prefix + '/')
        # Ensure it does not have an old-style multi version inheritance
        if self.build_parent_in_prefix and ":" in self.build_parent:
            raise BadConfigError(
                "Container {} has versioned build parent - it should be converted to just a name.".format(self.path),
            )
        # Load information from bay.yaml file
        if os.path.isfile(self.config_path):
            with open(self.config_path, "r") as fh:
                config_data = yaml.safe_load(fh.read()) or {}
        else:
            config_data = {}
        # Calculate links
        # TODO: Remove old, deprecated links format.
        self.links = {}
        config_links = config_data.get("links", {})
        if isinstance(config_links, list):
            # Old list format
            for link_name in config_links:
                self.links[link_name] = {"required": True}
        else:
            # New links format
            for link_name in (config_links.get("required") or []):
                self.links[link_name] = {"required": True}
            for link_name in (config_links.get("optional") or []):
                self.links[link_name] = {"required": False}
        # Old extra links key
        config_extra_links = config_data.get("extra_links", [])
        if config_extra_links:
            for link_name in config_extra_links:
                self.links[link_name] = {"required": False}
        # Parse waits from the config format
        self.waits = []
        for wait_dict in config_data.get("waits", []):
            for wait_type, params in wait_dict.items():
                if not isinstance(params, dict):
                    # TODO: Deprecate non-dictionary params
                    if wait_type == "time":
                        params = {"seconds": params}
                    else:
                        params = {"port": params}
                self.waits.append({"type": wait_type, "params": params})
        # Volumes is a dict of {container mountpoint: volume name/host path}
        self._bound_volumes = {}
        self._named_volumes = {}
        for mount_point, source in config_data.get("volumes", {}).items():
            # Old-style git link
            # TODO: Add warning here once we've converted enough of the dockerfiles
            git_match = self.git_volume_pattern.match(source)
            if git_match:
                source = "../{}/{}".format(git_match.group(1), git_match.group(2).lstrip("/"))
            # Split named volumes and directory mounts up
            if "/" in source:
                self._bound_volumes[mount_point] = os.path.abspath(os.path.join(self.graph.path, source))
            else:
                self._named_volumes[mount_point] = source
        # Volumes_mount is a deprecated key from the old buildable volumes system.
        # They turn into named volumes.
        # TODO: Deprecate volumes_mount
        for mount_point, source in config_data.get("volumes_mount", {}).items():
            self._named_volumes[mount_point] = source
        # Devmodes might also have git URLs
        self._devmodes = {}
        for name, mounts in config_data.get("devmodes", {}).items():
            # Allow for empty devmodes
            if not mounts:
                continue
            # Add each mount individually
            self._devmodes[name] = {}
            for mount_point, source in mounts.items():
                git_match = self.git_volume_pattern.match(source)
                if git_match:
                    source = "../{}/{}".format(git_match.group(1), git_match.group(2).lstrip("/"))
                self._devmodes[name][mount_point] = os.path.abspath(os.path.join(self.graph.path, source))
        # Ports is a dict of {port on container: host exposed port}
        self.ports = config_data.get("ports", {})
        # A list of checks to run before allowing a build (often for network connectivity)
        self.build_checks = config_data.get("build_checks", [])
        # If the container should launch into a foreground shell with its CMD when run, rather than
        # starting up in the background. Useful for test suites etc.
        self.foreground = config_data.get("foreground", False)
        # The image tag to use on the docker image. "local" is a special value that resolves to "latest" without
        # ever attempting to pull.
        self.image_tag = config_data.get("image_tag", "local")
        # Environment variables to send to the container
        self.environment = config_data.get("environment", {})
        # Fast kill says if the container is safe to kill immediately
        self.fast_kill = config_data.get("fast_kill", False)
        # System says if the container is a supporting "system" container, and lives and runs
        # outside of the profiles (e.g. it's ignored by bay restart, or bay up)
        self.system = config_data.get("system", False)
        # Abstract says if the container is not intended to ever be run or linked to, just
        # used as a base for other containers
        self.abstract = config_data.get("abstract", False)
        # Build args to pass into the container; right now, these are only settable by plugins.
        self.buildargs = {}
        # Store all extra data so plugins can get to it
        self.extra_data = {
            key: value
            for key, value in config_data.items()
            if key not in ["ports", "build_checks", "devmodes", "foreground", "links", "waits", "volumes", "image_tag"]
        }

    def get_parent_value(self, name, default):
        """
        Shortcut for getting inherited values from parents with a fallback.
        """
        if self.build_parent_in_prefix:
            return getattr(self.graph.build_parent(self), name)
        else:
            return default

    def get_ancestral_extra_data(self, key):
        """
        Returns a list of all extra data values with "key" from this container
        up through all build parents.
        """
        if self.build_parent_in_prefix:
            result = self.graph.build_parent(self).get_ancestral_extra_data(key)
            if key in self.extra_data:
                result.append(self.extra_data[key])
            return result
        elif key in self.extra_data:
            return [self.extra_data[key]]
        else:
            return []

    def get_named_volume_path(self, volume_name):
        """
        Returns the mount path, given a volume name.

        Raises an error if the volume_name provided is not moutned.
        """
        for path, name in self.named_volumes.items():
            if name == volume_name:
                return path
        raise ValueError("{} is not mounted".format(volume_name))

    @property
    def bound_volumes(self):
        value = self.get_parent_value("bound_volumes", {})
        value.update(self._bound_volumes)
        return value

    @property
    def named_volumes(self):
        value = self.get_parent_value("named_volumes", {})
        value.update(self._named_volumes)
        return value

    @property
    def devmodes(self):
        value = self.get_parent_value("devmodes", {})
        value.update(self._devmodes)
        return value
